import zenoh
import json
from zenoh import Session, Query, Sample
from typing import List, Tuple
import time

SLEEP = 1
MSG_COUNT = 1_000;
MSG_SIZE = [1_024, 131_072];


def open_session(endpoints: List[str]) -> Tuple[Session, Session]:
    # listen peer
    conf = zenoh.Config()
    conf.insert_json5("listen/endpoints", json.dumps(endpoints))
    conf.insert_json5("scouting/multicast/enabled", "false")
    print("[  ][01a] Opening peer01 session");
    peer01 = zenoh.open(conf)

    # connect peer
    conf = zenoh.Config()
    conf.insert_json5("connect/endpoints", json.dumps(endpoints))
    conf.insert_json5("scouting/multicast/enabled", "false")
    print("[  ][02a] Opening peer02 session");
    peer02 = zenoh.open(conf)

    return (peer01, peer02)


def close_session(peer01: Session, peer02: Session):
    print("[  ][01d] Closing peer01 session");
    peer01.close()
    print("[  ][02d] Closing peer02 session");
    peer02.close()


def run_session_qryrep(peer01: Session, peer02: Session):
    keyexpr = "test/session"

    for size in MSG_SIZE:
        num_requests = 0
        num_replies = 0
        num_errors = 0

        def queryable_callback(query: Query):
            nonlocal num_requests
            query.reply(Sample(keyexpr, bytes(size)))
            num_requests += 1

        print("[QR][01c] Queryable on peer01 session");
        queryable = peer01.declare_queryable(
            keyexpr,
            queryable_callback,
            complete=False
        )

        time.sleep(SLEEP)

        print(f"[QR][02c] Getting on peer02 session. {MSG_COUNT} msgs.");
        for _ in range(MSG_COUNT):
            replies = peer02.get(keyexpr, zenoh.Queue())
            for reply in replies.receiver:
                try:
                    unwraped_reply = reply.ok
                except:
                    unwraped_reply = None

                if unwraped_reply:
                    assert len(unwraped_reply.payload) == size
                    num_replies += 1
                else:
                    num_errors += 1

        time.sleep(SLEEP)
        print(f"[QR][02c] Got on peer02 session. {num_replies}/{MSG_COUNT} msgs.");
        assert num_replies == MSG_COUNT
        assert num_requests == MSG_COUNT
        assert num_errors == 0

        print("[QR][03c] Unqueryable on peer01 session");
        queryable.undeclare()


def test_session():
    zenoh.init_logger()
    (peer01, peer02) = open_session(["tcp/127.0.0.1:17447"])
    run_session_qryrep(peer01, peer02)
    close_session(peer01, peer02)



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
from .zenoh import _Encoding, _SampleKind, _CongestionControl, _Priority, _Reliability, _QueryTarget, _QueryConsolidation

class Priority(_Priority):
    """
    The priority of a sending operation.

    They are ordered à la Linux priority:
    ``Priority.REAL_TIME() < Priority.INTERACTIVE_HIGH() < Priority.INTERACTIVE_LOW() < Priority.DATA() < Priority.BACKGROUND()``
    """
    def __new__(cls, inner: _SampleKind):
        return super().__new__(cls, inner)
    @staticmethod
    def REAL_TIME() -> 'Priority':
        return Priority(_Priority.REAL_TIME)
    @staticmethod
    def INTERACTIVE_HIGH() -> 'Priority':
        return Priority(_Priority.INTERACTIVE_HIGH)
    @staticmethod
    def INTERACTIVE_LOW() -> 'Priority':
        return Priority(_Priority.INTERACTIVE_LOW)
    @staticmethod
    def DATA_HIGH() -> 'Priority':
        return Priority(_Priority.DATA_HIGH)
    @staticmethod
    def DATA() -> 'Priority':
        return Priority(_Priority.DATA)
    @staticmethod
    def DATA_LOW() -> 'Priority':
        return Priority(_Priority.DATA_LOW)
    @staticmethod
    def BACKGROUND() -> 'Priority':
        return Priority(_Priority.BACKGROUND)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __lt__(self, other) -> bool:
        return super().__lt__(other)
    def __le__(self, other) -> bool:
        return super().__le__(other)
    def __gt__(self, other) -> bool:
        return super().__gt__(other)
    def __ge__(self, other) -> bool:
        return super().__ge__(other)

class SampleKind(_SampleKind):
    "Similar to an HTTP METHOD: only PUT and DELETE are currently supported."
    def __new__(cls, inner: _SampleKind):
        return super().__new__(cls, inner)
    @staticmethod
    def PUT() -> 'SampleKind':
        return SampleKind(_SampleKind.PUT)
    @staticmethod
    def DELETE() -> 'SampleKind':
        return SampleKind(_SampleKind.DELETE)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

class CongestionControl(_CongestionControl):
    """
    Defines the network's behaviour regarding a message when heavily congested.
    """
    def __new__(cls, inner: _CongestionControl):
        return super().__new__(cls, inner)
    @staticmethod
    def DROP() -> 'CongestionControl':
        "Allows the message to be dropped if all buffers are full."
        return CongestionControl(_CongestionControl.DROP)
    @staticmethod
    def BLOCK() -> 'CongestionControl':
        """
        Prevents the message from being dropped at all cost.
        In the face of heavy congestion on a part of the network, this could result in your publisher node blocking.
        """
        return CongestionControl(_CongestionControl.BLOCK)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

class Encoding(_Encoding):
    def __new__(cls, inner: _Encoding):
        return super().__new__(cls, inner)
    @staticmethod
    def from_str(s: str) -> 'Encoding':
        return super(Encoding, Encoding).from_str(s)
    def append(self, s: str):
        super().append(s)
    @staticmethod
    def EMPTY() -> 'Encoding':
        return Encoding(_Encoding.EMPTY )
    @staticmethod
    def APP_OCTET_STREAM() -> 'Encoding':
        return Encoding(_Encoding.APP_OCTET_STREAM)
    @staticmethod
    def APP_CUSTOM() -> 'Encoding':
        return Encoding(_Encoding.APP_CUSTOM)
    @staticmethod
    def TEXT_PLAIN() -> 'Encoding':
        return Encoding(_Encoding.TEXT_PLAIN)
    @staticmethod
    def APP_PROPERTIES() -> 'Encoding':
        return Encoding(_Encoding.APP_PROPERTIES)
    @staticmethod
    def APP_JSON() -> 'Encoding':
        return Encoding(_Encoding.APP_JSON)
    @staticmethod
    def APP_SQL() -> 'Encoding':
        return Encoding(_Encoding.APP_SQL)
    @staticmethod
    def APP_INTEGER() -> 'Encoding':
        return Encoding(_Encoding.APP_INTEGER)
    @staticmethod
    def APP_FLOAT() -> 'Encoding':
        return Encoding(_Encoding.APP_FLOAT)
    @staticmethod
    def APP_XML() -> 'Encoding':
        return Encoding(_Encoding.APP_XML)
    @staticmethod
    def APP_XHTML_XML() -> 'Encoding':
        return Encoding(_Encoding.APP_XHTML_XML)
    @staticmethod
    def APP_X_WWW_FORM_URLENCODED() -> 'Encoding':
        return Encoding(_Encoding.APP_X_WWW_FORM_URLENCODED)
    @staticmethod
    def TEXT_JSON() -> 'Encoding':
        return Encoding(_Encoding.TEXT_JSON)
    @staticmethod
    def TEXT_HTML() -> 'Encoding':
        return Encoding(_Encoding.TEXT_HTML)
    @staticmethod
    def TEXT_XML() -> 'Encoding':
        return Encoding(_Encoding.TEXT_XML)
    @staticmethod
    def TEXT_CSS() -> 'Encoding':
        return Encoding(_Encoding.TEXT_CSS)
    @staticmethod
    def TEXT_CSV() -> 'Encoding':
        return Encoding(_Encoding.TEXT_CSV)
    @staticmethod
    def TEXT_JAVASCRIPT() -> 'Encoding':
        return Encoding(_Encoding.TEXT_JAVASCRIPT)
    @staticmethod
    def IMAGE_JPEG() -> 'Encoding':
        return Encoding(_Encoding.IMAGE_JPEG)
    @staticmethod
    def IMAGE_PNG() -> 'Encoding':
        return Encoding(_Encoding.IMAGE_PNG)
    @staticmethod
    def IMAGE_GIF() -> 'Encoding':
        return Encoding(_Encoding.IMAGE_GIF)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

class Reliability(_Reliability):
    "Used by subscribers to inform the network of the reliability it wishes to obtain."
    def __new__(cls, inner: _Reliability):
        return super().__new__(cls, inner)
    @staticmethod
    def BEST_EFFORT() -> 'CongestionControl':
        "Informs the network that dropping some messages is acceptable"
        return Reliability(_Reliability.BEST_EFFORT)
    @staticmethod
    def RELIABLE() -> 'CongestionControl':
        """
        Informs the network that this subscriber wishes for all publications to reliably reach it.

        Note that if a publisher puts a sample with the ``CongestionControl.DROP()`` option, this reliability
        requirement may be infringed to prevent slow readers from blocking the network.
        """
        return Reliability(_Reliability.RELIABLE)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

class QueryTarget(_QueryTarget):
    def __new__(cls, inner: _QueryTarget):
        return super().__new__(cls, inner)
    @staticmethod
    def BEST_MATCHING() -> 'QueryTarget':
        return QueryTarget(_QueryTarget.BEST_MATCHING)
    @staticmethod
    def ALL() -> 'QueryTarget':
        return QueryTarget(_QueryTarget.ALL)
    @staticmethod
    def ALL_COMPLETE() -> 'QueryTarget':
        return QueryTarget(_QueryTarget.ALL_COMPLETE)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

class QueryConsolidation(_QueryConsolidation):
    def __new__(cls, inner: _QueryConsolidation):
        return super().__new__(cls, inner)
    @staticmethod
    def AUTO() -> 'QueryConsolidation':
        return QueryConsolidation(_QueryConsolidation.AUTO)
    @staticmethod
    def NONE() -> 'QueryConsolidation':
        return QueryConsolidation(_QueryConsolidation.NONE)
    @staticmethod
    def MONOTONIC() -> 'QueryConsolidation':
        return QueryConsolidation(_QueryConsolidation.MONOTONIC)
    @staticmethod
    def LATEST() -> 'QueryConsolidation':
        return QueryConsolidation(_QueryConsolidation.LATEST)
    def __eq__(self, other) -> bool:
        return super().__eq__(other)
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#

import sys
import time
import argparse
import json
import zenoh
from zenoh import config, QueryTarget

# --- Command line argument parsing --- --- --- --- --- ---
parser = argparse.ArgumentParser(
    prog='z_get',
    description='zenoh get example')
parser.add_argument('--mode', '-m', dest='mode',
                    choices=['peer', 'client'],
                    type=str,
                    help='The zenoh session mode.')
parser.add_argument('--connect', '-e', dest='connect',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to connect to.')
parser.add_argument('--listen', '-l', dest='listen',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to listen on.')
parser.add_argument('--selector', '-s', dest='selector',
                    default='demo/example/**',
                    type=str,
                    help='The selection of resources to query.')
parser.add_argument('--target', '-t', dest='target',
                    choices=['ALL', 'BEST_MATCHING', 'ALL_COMPLETE', 'NONE'],
                    default='BEST_MATCHING',
                    type=str,
                    help='The target queryables of the query.')
parser.add_argument('--value', '-v', dest='value',
                    type=str,
                    help='An optional value to send in the query.')
parser.add_argument('--config', '-c', dest='config',
                    metavar='FILE',
                    type=str,
                    help='A configuration file.')

args = parser.parse_args()
conf = zenoh.Config.from_file(
    args.config) if args.config is not None else zenoh.Config()
if args.mode is not None:
    conf.insert_json5(zenoh.config.MODE_KEY, json.dumps(args.mode))
if args.connect is not None:
    conf.insert_json5(zenoh.config.CONNECT_KEY, json.dumps(args.connect))
if args.listen is not None:
    conf.insert_json5(zenoh.config.LISTEN_KEY, json.dumps(args.listen))
selector = args.selector
target = {
    'ALL': QueryTarget.ALL(),
    'BEST_MATCHING': QueryTarget.BEST_MATCHING(),
    'ALL_COMPLETE': QueryTarget.ALL_COMPLETE(),
}.get(args.target)

# Zenoh code  --- --- --- --- --- --- --- --- --- --- ---

# initiate logging
zenoh.init_logger()

print("Opening session...")
session = zenoh.open(conf)

print("Sending Query '{}'...".format(selector))
replies = session.get(selector, zenoh.Queue(), target=target, value=args.value, consolidation=zenoh.QueryConsolidation.NONE())
for reply in replies.receiver:
    try:
        print(">> Received ('{}': '{}')"
              .format(reply.ok.key_expr, reply.ok.payload.decode("utf-8")))
    except:
        print(">> Received (ERROR: '{}')"
              .format(reply.err.payload.decode("utf-8")))


session.close()



# Copyright (c) 2017, 2022 ZettaScale Technology Inc.

# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.

# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0

# Contributors:
#   ZettaScale Zenoh team, <zenoh@zettascale.tech>
# 
from os import path
from subprocess import Popen, PIPE
import time

examples = path.dirname(path.realpath(__file__)).replace('tests', 'examples')
tab = "\t"
ret = "\r\n"
KILL = -9

class Pyrun:
	def __init__(self, p, args=None) -> None:
		if args is None:
			args = []
		self.name = p
		print(f"starting {self.name}")
		self.process: Popen = Popen(["python3", path.join(examples, p), *args], stdin=PIPE, stdout=PIPE, stderr=PIPE)
		self.start = time.time()
		self.end = None
		self._stdouts = []
		self._stderrs = []
	def dbg(self):
		self.wait()
		print(f"{self.name} stdout:")
		print(f"{tab}{tab.join(self.stdout)}")
		print(f"{self.name} stderr:")
		print(f"{tab}{tab.join(self.stderr)}")
	def status(self, expecting=0, do_print=True):
		status = self.wait()
		formatted = f"{self.name}: returned {status} (expected {expecting}) - {self.time:.2}s"
		if do_print:
			print(formatted)
		return formatted if status != expecting else None
	def wait(self):
		try:
			code = self.process.wait(timeout=10)
		except:
			self.process.kill()
			code = self.process.wait(timeout=10)
		if self.end is None:
			self.end = time.time()
		return code
	@property
	def stdout(self):
		self._stdouts.extend(line.decode('utf8') for line in self.process.stdout.readlines())
		return self._stdouts
	@property
	def stderr(self):
		self._stderrs.extend(line.decode('utf8') for line in self.process.stderr.readlines())
		return self._stderrs
	@property
	def time(self):
		return None if self.end is None else (self.end - self.start)

errors = []

info = Pyrun("z_info.py")
if info.status():
	info.dbg()
	errors.append(info.status())
scout = Pyrun("z_scout.py")
if scout.status():
	scout.dbg()
	errors.append(scout.status())

storage = Pyrun("z_storage.py")
sub = Pyrun("z_sub.py")
pull = Pyrun("z_pull.py")
time.sleep(1)
put = Pyrun("z_put.py")
if put.status():
	put.dbg()
	errors.append(put.status())
time.sleep(1)
pub = Pyrun("z_pub.py", ["--iter=2"])
time.sleep(4)
try:
	pull.process.stdin.write(b"\n")
	pull.process.stdin.flush()
	time.sleep(1)
	pull.process.stdin.write(b"q\n")
	pull.process.stdin.flush()
	pull.process.stdin.close()
except Exception as e:
	pull.dbg()
	errors.append(f"pull stdin sequence failed: {e}")
if pub.status():
	pub.dbg()
	errors.append(pub.status())
if pull.status():
	pull.dbg()
	errors.append(pull.status())
subout = "".join(pull.stdout)
if not ("Received PUT ('demo/example/zenoh-python-put': 'Put from Python!')" in subout):
	errors.append("z_pull didn't catch put")
if not ("Received PUT ('demo/example/zenoh-python-pub': '[   1] Pub from Python!')" in subout):
	errors.append("z_pull didn't catch second z_pub")
if any(("z_pull" in error) for error in errors):
	pull.dbg()

queryable = Pyrun("z_queryable.py", ["-k=demo/example/zenoh-python-queryable"])
time.sleep(1)
get = Pyrun("z_get.py", ["-s=demo/example/zenoh-python-queryable"])
if get.status():
	get.dbg()
	errors.append(get.status())
if not ("Received ('demo/example/zenoh-python-queryable': 'Queryable from Python!')" in "".join(get.stdout)):
	get.dbg()
	queryable.dbg()
	errors.append("z_get didn't get a response from z_queryable")

try:
	queryable.process.stdin.write(b"q\n")
	queryable.process.stdin.flush()
	queryable.process.stdin.close()
except Exception as e:
	errors.append(f"queryable stdin sequence failed: {e}")
if queryable.status():
	queryable.dbg()
	errors.append(queryable.status())
queryableout = "".join(queryable.stdout)
if not ("Received Query 'demo/example/zenoh-python-queryable'" in queryableout):
	errors.append("z_queryable didn't catch query")
if any(("z_queryable" in error) for error in errors):
	queryable.dbg()

time.sleep(1)
get = Pyrun("z_get.py", ["-s=demo/example/zenoh-python-put"])
if get.status():
	get.dbg()
	errors.append(get.status())
if not ("Received ('demo/example/zenoh-python-put': 'Put from Python!')" in "".join(get.stdout)):
	get.dbg()
	errors.append("z_get didn't get a response from z_storage about put")
if any(("z_get" in error) for error in errors):
	get.dbg()

time.sleep(1)
delete = Pyrun("z_delete.py")
if delete.status():
	delete.dbg()
	errors.append(delete.status())

time.sleep(1)
get = Pyrun("z_get.py", ["-s=demo/example/zenoh-python-put"])
if get.status():
	get.dbg()
	errors.append(get.status())
if ("Received ('demo/example/zenoh-python-put': 'Put from Python!')" in "".join(get.stdout)):
	# storage.dbg()
	errors.append("z_get did get a response from z_storage about put after delete")
if any(("z_get" in error) for error in errors):
	get.dbg()

try:
	sub.process.stdin.write(b"q\n")
	sub.process.stdin.flush()
	sub.process.stdin.close()
except Exception as e:
	errors.append(f"pub stdin sequence failed: {e}")
if sub.status():
	sub.dbg()
	errors.append(sub.status())
subout = "".join(sub.stdout)
if not ("Received PUT ('demo/example/zenoh-python-put': 'Put from Python!')" in subout):
	errors.append("z_sub didn't catch put")
if not ("Received PUT ('demo/example/zenoh-python-pub': '[   1] Pub from Python!')" in subout):
	errors.append("z_sub didn't catch second z_pub")
if not ("Received DELETE ('demo/example/zenoh-python-put': '')" in subout):
	errors.append("z_sub didn't catch delete")
if any(("z_sub" in error) for error in errors):
	sub.dbg()

try:
	storage.process.stdin.write(b"q\n")
	storage.process.stdin.flush()
	storage.process.stdin.close()
except Exception as e:
	errors.append(f"storage stdin sequence failed: {e}")
if storage.status():
	storage.dbg()
	errors.append(storage.status())
storageout = "".join(storage.stdout)
if not ("Received PUT ('demo/example/zenoh-python-put': 'Put from Python!')" in storageout):
	errors.append("z_storage didn't catch put")
if not ("Received PUT ('demo/example/zenoh-python-pub': '[   1] Pub from Python!')" in storageout):
	errors.append("z_storage didn't catch second z_pub")
if not ("Received DELETE ('demo/example/zenoh-python-put': '')" in storageout):
	errors.append("z_storage didn't catch delete")
if not ("Received Query 'demo/example/zenoh-python-put'" in storageout):
	errors.append("z_storage didn't catch query")
if any(("z_storage" in error) for error in errors):
	storage.dbg()

sub_thr = Pyrun("z_sub_thr.py")
pub_thr = Pyrun("z_pub_thr.py", ["128"])
time.sleep(5)
sub_thr.process.kill()
pub_thr.process.kill()
if sub_thr.status(KILL):
	sub_thr.dbg()
	errors.append(sub_thr.status())
if pub_thr.status(KILL):
	pub_thr.dbg()
	errors.append(pub_thr.status())


if len(errors):
	message = f"Found {len(errors)} errors: {(ret+tab) + (ret+tab).join(errors)}"
	raise Exception(message)


#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
from typing import Union, Dict
from .zenoh import _KeyExpr, _Selector

IntoKeyExpr = Union['KeyExpr', _KeyExpr, str]

class KeyExpr(_KeyExpr):
    """
    Zenoh's address space is designed around keys which serve as the names of ressources.

    Keys are slash-separated lists of non-empty UTF8 strings. They may not contain the following characters: ``$*#?``.
    
    Zenoh's operations are executed on key expressions, a small language that allows the definition
    of sets of keys via the use of wildcards:

     - ``*`` is the single-chunk wildcard, and will match any chunk: ``a/*/c`` will match ``a/b/c``, ``a/hello/c``, etc...
     - ``**`` is the 0 or more chunks wildcard: ``a/**/c`` matches ``a/c``, ``a/b/c``, ``a/b/hello/c``, etc...
     - ``$*`` is the subchunk wildcard, it will match any amount of non-/ characters: ``a/b$*`` matches ``a/b``, ``a/because``, ``a/blue``... but not ``a/c`` nor ``a/blue/c``
    
    To allow for better performance and gain the property that two key expressions define the same
    set if and only if they are the same string, the rules of canon form are mandatory for a key
    expression to be propagated by a Zenoh network:

     - ``**/**`` may not exist, as it could always be replaced by the shorter ``**``,
     - ``**/*`` may not exist, and must be written as its equivalent ``*/**`` instead,
     - ``$*`` may not exist alone in a chunk, as it must be written ``*`` instead.

    The ``KeyExpr.autocanonize`` constructor exists to correct eventual infrigements of the canonization rules.

    A KeyExpr is a string that has been validated to be a valid Key Expression.
    """
    def __new__(cls, expr: IntoKeyExpr):
        """
        The default constructor for KeyExpr will ensure that the passed expression is valid.
        It won't however try to correct expressions that aren't canon.

        You may use ``KeyExpr.autocanonize(expr)`` instead if you are unsure if the expression
        you will use for construction will be canon.

        Raises a zenoh.ZError exception if ``expr`` is not a valid key expression.
        """
        if isinstance(expr, KeyExpr):
            return expr
        elif isinstance(expr, _KeyExpr):
            return _KeyExpr.__new__(cls, expr)
        else:
            return _KeyExpr.__new__(cls, _KeyExpr.new(expr))
    
    def _upgrade_(this: _KeyExpr) -> 'KeyExpr':
        return _KeyExpr.__new__(KeyExpr, expr)

    @staticmethod
    def autocanonize(expr: str) -> 'KeyExpr':
        """
        This alternative constructor for key expressions will attempt to canonize the passed
        expression before checking if it is valid.

        Raises a zenoh.ZError exception if ``expr`` is not a valid key expression.
        """
        if isinstance(expr, KeyExpr):
            return expr
        else:
            e = _KeyExpr.autocanonize(expr)
            return KeyExpr(e.as_str())
    
    def intersects(self, other: IntoKeyExpr) -> bool:
        """
        This method returns ``True`` if there exists at least one key that belongs to both sets
        defined by ``self`` and ``other``. 
        """
        return super().intersects(KeyExpr(other))
    
    def includes(self, other: IntoKeyExpr) -> bool:
        """
        This method returns ``True`` if all of the keys defined by ``other`` also belong to the set
        defined by ``self``.
        """
        return super().includes(KeyExpr(other))
    
    def undeclare(self, session: 'Session'):
        """
        Undeclares a key expression previously declared on the session.
        """
        super().undeclare(session)
    
    def __eq__(self, other: IntoKeyExpr) -> bool:
        """
        Corresponds to set equality.
        """
        return super().__eq__(KeyExpr(other))
    
    def __truediv__(self, other: IntoKeyExpr) -> 'KeyExpr':
        """
        Joins two key expressions with a ``/``.

        Raises a zenoh.ZError exception if ``other`` is not a valid key expression.
        """
        return KeyExpr.autocanonize(f"{self}/{other}")
    
    def __str__(self):
        return super().__str__()
    
    def __hash__(self):
        return super().__hash__()

IntoSelector = Union['Selector', _Selector, IntoKeyExpr]
class Selector(_Selector):
    """
    A selector is the combination of a [Key Expression](crate::prelude::KeyExpr), which defines the
    set of keys that are relevant to an operation, and a ``parameters``, a set of key-value pairs
    with a few uses:

     * specifying arguments to a queryable, allowing the passing of Remote Procedure Call parameters
     * filtering by value,
     * filtering by metadata, such as the timestamp of a value,

    When in string form, selectors look a lot like a URI, with similar semantics:

     * the ``key_expr`` before the first ``?`` must be a valid key expression.
     * the ``parameters`` after the first ``?`` should be encoded like the query section of a URL:

        * key-value pairs are separated by ``&``,
        * the key and value are separated by the first ``=``,
        * in the absence of ``=``, the value is considered to be the empty string,
        * both key and value should use percent-encoding to escape characters,
        * defining a value for the same key twice is considered undefined behavior.

    Zenoh intends to standardize the usage of a set of keys. To avoid conflicting with RPC parameters,
    the Zenoh team has settled on reserving the set of keys that start with non-alphanumeric characters.

    This document will summarize the standardized keys for which Zenoh provides helpers to facilitate
    coherent behavior for some operations.

    Queryable implementers are encouraged to prefer these standardized keys when implementing their
    associated features, and to prefix their own keys to avoid having conflicting keys with other
    queryables.

    Here are the currently standardized keys for Zenoh:

     * ``_time``: used to express interest in only values dated within a certain time range, values for
       this key must be readable by the [Zenoh Time DSL](zenoh_util::time_range::TimeRange) for the value to be considered valid.
     * ``_filter``: *TBD* Zenoh intends to provide helper tools to allow the value associated with
       this key to be treated as a predicate that the value should fulfill before being returned.
       A DSL will be designed by the Zenoh team to express these predicates.
    """
    def __new__(cls, selector: IntoSelector):
        if isinstance(selector, Selector):
            return selector
        if isinstance(selector, _Selector):
            return Selector._upgrade_(selector)
        return Selector._upgrade_(super().new(str(selector)))
    @staticmethod
    def _upgrade_(this: _Selector) -> 'Selector':
        return _Selector.__new__(Selector, this)
    @property
    def key_expr(self) -> KeyExpr:
        "The key expression part of the selector."
        return KeyExpr(super().key_expr)
    @property
    def parameters(self):
        "The value selector part of the selector."
        return super().parameters
    @parameters.setter
    def set_parameters(self, parameters: str):
        super().parameters = parameters
    def decode_parameters(self) -> Dict[str, str]:
        """
        Decodes the value selector part of the selector.

        Raises a ZError if some keys were duplicated: duplicated keys are considered undefined behaviour,
        but we encourage you to refuse to process incoming messages with duplicated keys, as they might be
        attempting to use HTTP Parameter Pollution like exploits.
        """
        return super().decode_parameters()
    def __str__(self):
        return super().__str__()


#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
from typing import Union
from .zenoh import _Config
import json

class Config(_Config):
    def __init__(self):
        super().__init__()
    @staticmethod
    def from_file(filename: str):
        """
        Reads the configuration from a file.
        The file's extension must be json, json5 or yaml.
        """
        c = super(Config, Config).from_file(filename)
        return c
    @staticmethod
    def from_obj(obj):
        """
        Reads the configuration from ``obj`` as if it was a JSON file.
        """
        c = Config.from_json5(json.dumps(obj))
        return c
    @staticmethod
    def from_json5(json: str):
        """
        Reads the configuration from a JSON5 string.

        JSON5 is a superset of JSON, so any JSON string is a valid input for this function.
        """
        c =  super(Config, Config).from_json5(json)
        return c
    
    def get_json(self, path: str) -> str:
        """
        Returns the part of the configuration at ``path``,
        in a JSON-serialized form.
        """
        return super().get_json(path)
    
    def insert_json5(self, path: str, value: str) -> str:
        """
        Inserts the provided value (read from JSON) at the given path in the configuration.
        """
        return super().insert_json5(path, value)

MODE_KEY = "mode"
CONNECT_KEY = "connect/endpoints"
LISTEN_KEY = "listen/endpoints"


#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#

import zenoh

# initiate logging
zenoh.init_logger()

print("Scouting...")
scout = zenoh.scout(what = "peer|router", timeout=1.0)

for hello in scout.receiver():
    print(hello)



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
from typing import Dict, Optional

from .zenoh import _Query, _Queryable
from .keyexpr import KeyExpr, Selector
from .value import Sample, Value, IntoValue, IntoSample

class Queryable:
    """
    A handle to a queryable.
    
    Its main purpose is to keep the queryable active as long as it exists.

    When constructed through ``Session.declare_queryable(session, keyexpr, handler)``, it exposes ``handler``'s receiver
    through ``self.receiver``.
    """
    def __init__(self, inner: _Queryable, receiver):
        self._inner_ = inner
        self.receiver = receiver
    
    def undeclare(self):
        "Stops the queryable."
        self._inner_ = None

class Query(_Query):
    def __new__(cls, inner: _Query):
        return super().__new__(cls, inner)
    @property
    def key_expr(self) -> KeyExpr:
        "The query's targeted key expression"
        return KeyExpr(super().key_expr)
    @property
    def parameters(self) -> str:
        """
        The query's value selector.
        If you'd rather not bother with parsing it yourself, use ``self.decode_parameters()`` instead.
        """
        return super().parameters
        
    def decode_parameters(self) -> Dict[str, str]:
        """
        Decodes the value selector into a dictionary.

        Raises a ZError if duplicate keys are found, as they might otherwise be used for HTTP Parameter Pollution like attacks.
        """
        return super().decode_parameters()
    @property
    def selector(self) -> Selector:
        """
        The query's selector as a whole.
        """
        return Selector._upgrade_(super().selector)
    @property
    def value(self) -> Optional[Value]:
        """
        The query's value.
        """
        return Value._upgrade_(super().value)
    def reply(self, sample: Sample):
        """
        Allows you to reply to a query.
        You may send any amount of replies to a single query, including 0.
        """
        super().reply(sample)
    def reply_err(self, value: IntoValue):
        """
        Allows you to reply to a query with an error.
        You may send any amount of replies to a single query, including 0.
        Sending error responses does not exclude sending other responses.
        """
        super().reply_err(Value(value))



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
from .zenoh import init_logger, scout as _scout
from .keyexpr import IntoKeyExpr, IntoSelector, KeyExpr, Selector
from .config import Config
from .session import Session, Publisher, Subscriber, PullSubscriber, Info
from .enums import CongestionControl, Encoding, Priority, QueryConsolidation, QueryTarget, Reliability, SampleKind
from .value import Hello, Value, IntoValue, IValue, Sample, IntoSample, ZenohId, Timestamp, Reply
from .closures import Closure, IClosure, IntoClosure, Handler, IHandler, IntoHandler, ListCollector, Queue
from .queryable import Queryable, Query
from typing import Any

def open(*args, **kwargs):
    """
    Open a Zenoh session.

    :param config: The configuration of the Zenoh session
    :type config: Config
    :rtype: Session

    :Example:

    >>> import zenoh
    >>> s = zenoh.open(zenoh.Config())
    """
    return Session(*args, **kwargs)

class Scout:
    def __init__(self, inner, receiver):
        self._inner_ = inner
        self.receiver = receiver
    
    def stop(self):
        self._inner_ = None

def scout(handler: IntoHandler[Hello, Any, Any] = None, what: str = None, config: Config = None, timeout=None):
    """
    Scout for routers and/or peers.

    This spawns a task that periodically sends scout messages for a specified duration and returns
    a list of received :class:`Hello` messages.

    :param what: The kind of zenoh process to scout for
    :param config: The configuration to use for scouting
    :param timeout: the duration of scout (in seconds)
    :param handler:
    :rtype: list of :class:`Hello`

    :Example:

    >>> import zenoh
    >>> for hello in zenoh.scout(what = "peer|router", timeout=1.0).receiver():
    ...     print(hello)
    """
    from threading import Timer
    if handler is None:
        handler = ListCollector()
    handler = Handler(handler, lambda x: Hello._upgrade_(x))
    scout = _scout(handler.closure, config, what)
    scout = Scout(scout, handler.receiver)
    if timeout:
        Timer(timeout, lambda: scout.stop()).start()
    return scout



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#

import sys
import time
from datetime import datetime
import argparse
import json
import zenoh
from threading import Thread
from zenoh import Reliability, Sample

# --- Command line argument parsing --- --- --- --- --- ---
parser = argparse.ArgumentParser(
    prog='z_sub',
    description='zenoh sub example')
parser.add_argument('--mode', '-m', dest='mode',
                    choices=['peer', 'client'],
                    type=str,
                    help='The zenoh session mode.')
parser.add_argument('--connect', '-e', dest='connect',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to connect to.')
parser.add_argument('--listen', '-l', dest='listen',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to listen on.')
parser.add_argument('--key', '-k', dest='key',
                    default='demo/example/**',
                    type=str,
                    help='The key expression to subscribe to.')
parser.add_argument('--config', '-c', dest='config',
                    metavar='FILE',
                    type=str,
                    help='A configuration file.')

args = parser.parse_args()
conf = zenoh.Config.from_file(
    args.config) if args.config is not None else zenoh.Config()
if args.mode is not None:
    conf.insert_json5(zenoh.config.MODE_KEY, json.dumps(args.mode))
if args.connect is not None:
    conf.insert_json5(zenoh.config.CONNECT_KEY, json.dumps(args.connect))
if args.listen is not None:
    conf.insert_json5(zenoh.config.LISTEN_KEY, json.dumps(args.listen))
key = args.key

# Zenoh code  --- --- --- --- --- --- --- --- --- --- ---



# initiate logging
zenoh.init_logger()

print("Opening session...")
session = zenoh.open(conf)

print("Declaring Subscriber on '{}'...".format(key))


# WARNING, you MUST store the return value in order for the subscription to work!!
# This is because if you don't, the reference counter will reach 0 and the subscription
# will be immediately undeclared.
sub = session.declare_subscriber(key, zenoh.Queue(), reliability=Reliability.RELIABLE())

def consumer():
    for sample in sub.receiver: # zenoh.Queue's receiver (the queue itself) is an iterator
        print(f">> [Subscriber] Received {sample.kind} ('{sample.key_expr}': '{sample.payload.decode('utf-8')}')")

t = Thread(target=consumer)
t.start()
print("Enter 'q' to quit...")
c = '\0'
while c != 'q':
    c = sys.stdin.read(1)
    if c == '':
        time.sleep(1)

# Cleanup: note that even if you forget it, cleanup will happen automatically when 
# the reference counter reaches 0
sub.undeclare()
t.join()
session.close()



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#

import sys
import time
import argparse
import json
import zenoh
from zenoh import config, Value

# --- Command line argument parsing --- --- --- --- --- ---
parser = argparse.ArgumentParser(
    prog='z_put',
    description='zenoh put example')
parser.add_argument('--mode', '-m', dest='mode',
                    choices=['peer', 'client'],
                    type=str,
                    help='The zenoh session mode.')
parser.add_argument('--connect', '-e', dest='connect',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to connect to.')
parser.add_argument('--listen', '-l', dest='listen',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to listen on.')
parser.add_argument('--key', '-k', dest='key',
                    default='demo/example/zenoh-python-put',
                    type=str,
                    help='The key expression to write.')
parser.add_argument('--value', '-v', dest='value',
                    default='Put from Python!',
                    type=str,
                    help='The value to write.')
parser.add_argument('--config', '-c', dest='config',
                    metavar='FILE',
                    type=str,
                    help='A configuration file.')

args = parser.parse_args()
conf = zenoh.Config.from_file(
    args.config) if args.config is not None else zenoh.Config()
if args.mode is not None:
    conf.insert_json5(zenoh.config.MODE_KEY, json.dumps(args.mode))
if args.connect is not None:
    conf.insert_json5(zenoh.config.CONNECT_KEY, json.dumps(args.connect))
if args.listen is not None:
    conf.insert_json5(zenoh.config.LISTEN_KEY, json.dumps(args.listen))
key = args.key
value = args.value

# Zenoh code  --- --- --- --- --- --- --- --- --- --- ---

# initiate logging
zenoh.init_logger()

print("Opening session...")
session = zenoh.open(conf)

print("Putting Data ('{}': '{}')...".format(key, value))
session.put(key, value)

# --- Examples of put with other types:

# - Integer
# session.put('/demo/example/Integer', 3)

# - Float
# session.put('/demo/example/Float', 3.14)

# - Properties (as a Dictionary with str only)
# session.put('/demo/example/Properties', {'p1': 'v1', 'p2': 'v2'})

# - Json (str format)
# session.put('/demo/example/Json',
#             json.dumps(['foo', {'bar': ('baz', None, 1.0, 2)}]),
#             encoding=Encoding.TEXT_JSON)

# - Raw ('application/octet-stream' encoding by default)
# session.put('/demo/example/Raw', b'\x48\x69\x21')

# - Custom encoding
# session.put('/demo/example/Custom',
#             b'\x48\x69\x21',
#             encoding='my_encoding')

# - UTF-16 String specifying the charset as Encoding suffix
# session.put('/demo/example/UTF-16',
#             'hello'.encode('utf-16'),
#             encoding=Encoding.TEXT_PLAIN.with_suffix(';charset=utf-16'))

session.close()



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#

import sys
import time
import argparse
import json
import zenoh
from zenoh import Reliability

# --- Command line argument parsing --- --- --- --- --- ---
parser = argparse.ArgumentParser(
    prog='z_sub_thr',
    description='zenoh throughput sub example')
parser.add_argument('--mode', '-m', dest='mode',
                    choices=['peer', 'client'],
                    type=str,
                    help='The zenoh session mode.')
parser.add_argument('--connect', '-e', dest='connect',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to connect to.')
parser.add_argument('--listen', '-l', dest='listen',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to listen on.')
parser.add_argument('--number', '-n', dest='number',
                    default=50000,
                    metavar='NUMBER',
                    action='append',
                    type=int,
                    help='Number of messages in each throughput measurements.')
parser.add_argument('--config', '-c', dest='config',
                    metavar='FILE',
                    type=str,
                    help='A configuration file.')

args = parser.parse_args()
conf = zenoh.Config.from_file(args.config) if args.config is not None else zenoh.Config()
if args.mode is not None:
    conf.insert_json5(zenoh.config.MODE_KEY, json.dumps(args.mode))
if args.connect is not None:
    conf.insert_json5(zenoh.config.CONNECT_KEY, json.dumps(args.connect))
if args.listen is not None:
    conf.insert_json5(zenoh.config.LISTEN_KEY, json.dumps(args.listen))
n = args.number



batch_count = 0
count = 0
start = None
global_start = None

def listener(sample):
    global n, count, batch_count, start, global_start
    if count == 0:
        start = time.time()
        if global_start is None:
            global_start = start
        count += 1
    elif count < n:
        count += 1
    else:
        stop = time.time()
        print(f"{n / (stop - start):.6f} msgs/sec")
        batch_count += 1
        count = 0

def report():
    global n, m, count, batch_count,  global_start
    end = time.time()
    total = batch_count * n + count
    print(f"Received {total} messages in {end - global_start}: averaged {total / (end - global_start):.6f} msgs/sec")

# initiate logging
zenoh.init_logger()

session = zenoh.open(conf)

# By explicitly constructing the `Closure`, the `Queue` that's normally inserted between the callback and zenoh is removed.
# Only do this if your callback runs faster than the minimum expected delay between two samples.
sub = session.declare_subscriber("test/thr", zenoh.Closure((listener, report)), reliability=Reliability.RELIABLE())

print("Enter 'q' to quit...")
c = '\0'
while c != 'q':
    c = sys.stdin.read(1)
    if c == '':
        time.sleep(1)

sub.undeclare()
session.close()
# while `sub.undeclare()` only returns once the unsubscription is done (no more callbacks will be queued from that instant), already queued callbacks may still be running in threads that Python can't see.
time.sleep(0.1)


#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#

import sys
import time
import argparse
import json
import zenoh
from zenoh import Reliability, SampleKind, Query, Sample, KeyExpr

# --- Command line argument parsing --- --- --- --- --- ---
parser = argparse.ArgumentParser(
    prog='z_storage',
    description='zenoh storage example')
parser.add_argument('--mode', '-m', dest='mode',
                    choices=['peer', 'client'],
                    type=str,
                    help='The zenoh session mode.')
parser.add_argument('--connect', '-e', dest='connect',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to connect to.')
parser.add_argument('--listen', '-l', dest='listen',
                    metavar='ENDPOINT',
                    action='append',
                    type=str,
                    help='Endpoints to listen on.')
parser.add_argument('--key', '-k', dest='key',
                    default='demo/example/**',
                    type=str,
                    help='The key expression matching resources to store.')
parser.add_argument('--complete', dest='complete',
                    default=False,
                    action='store_true',
                    help='Declare the storage as complete w.r.t. the key expression.')
parser.add_argument('--config', '-c', dest='config',
                    metavar='FILE',
                    type=str,
                    help='A configuration file.')

args = parser.parse_args()
conf = zenoh.Config.from_file(
    args.config) if args.config is not None else zenoh.Config()
if args.mode is not None:
    conf.insert_json5(zenoh.config.MODE_KEY, json.dumps(args.mode))
if args.connect is not None:
    conf.insert_json5(zenoh.config.CONNECT_KEY, json.dumps(args.connect))
if args.listen is not None:
    conf.insert_json5(zenoh.config.LISTEN_KEY, json.dumps(args.listen))
key = args.key
complete = args.complete

# Zenoh code  --- --- --- --- --- --- --- --- --- --- ---

store = {}


def listener(sample: Sample):
    print(">> [Subscriber] Received {} ('{}': '{}')"
          .format(sample.kind, sample.key_expr, sample.payload.decode("utf-8")))
    if sample.kind == SampleKind.DELETE():
        store.pop(sample.key_expr, None)
    else:
        store[sample.key_expr] = sample


def query_handler(query: Query):
    print(">> [Queryable ] Received Query '{}'".format(query.selector))
    replies = []
    for stored_name, sample in store.items():
        if query.key_expr.intersects(stored_name):
            query.reply(sample)


# initiate logging
zenoh.init_logger()

print("Opening session...")
session = zenoh.open(conf)

print("Declaring Subscriber on '{}'...".format(key))
sub = session.declare_subscriber(
    key, listener, reliability=Reliability.RELIABLE())

print("Declaring Queryable on '{}'...".format(key))
queryable = session.declare_queryable(key, query_handler, complete)

print("Enter 'q' to quit...")
c = '\0'
while c != 'q':
    c = sys.stdin.read(1)
    if c == '':
        time.sleep(1)

sub.undeclare()
queryable.undeclare()
session.close()



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
from typing import Union, Any, List

from .zenoh import _Session, _Config, _Publisher, _Subscriber, _PullSubscriber

from .keyexpr import KeyExpr, IntoKeyExpr, Selector, IntoSelector
from .config import Config
from .closures import IntoHandler, Handler, Receiver
from .enums import *
from .value import IntoValue, Value, Sample, Reply, ZenohId
from .queryable import Queryable, Query


class Publisher:
    "Use ``Publisher`` (constructed with ``Session.declare_publisher``) when you want to send values often for the same key expression, as declaring them informs Zenoh that this is you intent, and optimizations will be set up to do so."

    def __init__(self, p: _Publisher):
        self._inner_ = p

    def put(self, value: IntoValue, encoding: Encoding = None):
        "An optimised version of ``session.put(self.key_expr, value, encoding=encoding)``"
        self._inner_.put(Value(value, encoding))

    def delete(self):
        "An optimised version of ``session.delete(self.key_expr)``"
        self._inner_.delete()

    @property
    def key_expr(self) -> KeyExpr:
        "This ``Publisher``'s key expression"
        return KeyExpr(self._inner_.key_expr)

    def undeclare(self):
        "Stops the publisher."
        self._inner_ = None


class Subscriber:
    """
    A handle to a subscription.

    Its main purpose is to keep the subscription active as long as it exists.

    When constructed through ``Session.declare_subscriber(session, keyexpr, handler)``, it exposes ``handler``'s receiver
    through ``self.receiver``.
    """

    def __init__(self, s: _Subscriber, receiver=None):
        self._subscriber_ = s
        self.receiver = receiver

    def undeclare(self):
        "Undeclares the subscription"
        self._subscriber_ = None


class PullSubscriber:
    """
    A handle to a pull subscription.

    Its main purpose is to keep the subscription active as long as it exists.

    When constructed through ``Session.declare_pull_subscriber(session, keyexpr, handler)``, it exposes ``handler``'s receiver
    through ``self.receiver``.

    Calling ``self.pull()`` will prompt the Zenoh network to send a new sample when available.
    """

    def __init__(self, s: _PullSubscriber, receiver=None):
        self._subscriber_ = s
        self.receiver = receiver

    def pull(self):
        """
        Prompts the Zenoh network to send a new sample if available.
        Note that this sample will not be returned by this function, but provided to the handler's callback.
        """
        self._subscriber_.pull()

    def undeclare(self):
        "Undeclares the subscription"
        self._subscriber_ = None


class Session(_Session):
    """
    A Zenoh Session, the core interraction point with a Zenoh network.

    Note that most applications will only need a single instance of ``Session``. You should _never_ construct one session per publisher/subscriber, as this will significantly increase the size of your Zenoh network, while preventing potential locality-based optimizations.
    """
    def __new__(cls, config: Union[Config, Any] = None):
        if config is None:
            return super().__new__(cls)
        elif isinstance(config, _Config):
            return super().__new__(cls, config)
        else:
            return super().__new__(cls, Config.from_obj(config))

    def put(self, keyexpr: IntoKeyExpr, value: IntoValue, encoding=None,
            priority: Priority = None, congestion_control: CongestionControl = None,
            sample_kind: SampleKind = None):
        """
        Sends a value over Zenoh.

        Subscribers on an expression that intersect with ``keyexpr`` will receive the sample.
        Storages will store the value if ``keyexpr`` is non-wild, or update the values for all known keys that are included in ``keyexpr`` if it is wild.

        :param keyexpr: The key expression to publish
        :param value: The value to send
        :param priority: The priority to use when routing the published data
        :param congestion_control: The congestion control to use when routing the published data
        :param sample_kind: The kind of sample to send

        :Examples:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> s.put('key/expression', 'value')
        """
        value = Value(value, encoding)
        keyexpr = KeyExpr(keyexpr)
        kwargs = dict()
        if priority is not None:
            kwargs['priority'] = priority
        if congestion_control is not None:
            kwargs['congestion_control'] = congestion_control
        if sample_kind is not None:
            kwargs['sample_kind'] = sample_kind
        return super().put(keyexpr, value, **kwargs)

    def config(self) -> Config:
        """Returns a configuration object that can be used to alter the session's configuration at runtime.

        Note that in Python specifically, the config you passed to the session becomes the result of this
        function if you passed one, letting you keep using it.
        """
        return super().config()

    def delete(self, keyexpr: IntoKeyExpr,
               priority: Priority = None, congestion_control: CongestionControl = None):
        """
        Deletes the values associated with the keys included in ``keyexpr``.
        
        This uses the same mechanisms as ``session.put``, and will be received by subscribers.
        This operation is especially useful with storages.

        :param keyexpr: The key expression to publish
        :param priority: The priority to use when routing the delete
        :param congestion_control: The congestion control to use when routing the delete

        :Examples:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> s.delete('key/expression')
        """
        keyexpr = KeyExpr(keyexpr)
        kwargs = dict()
        if priority is not None:
            kwargs['priority'] = priority
        if congestion_control is not None:
            kwargs['congestion_control'] = congestion_control
        return super().delete(keyexpr, **kwargs)

    def get(self, selector: IntoSelector, handler: IntoHandler[Reply, Any, Receiver], consolidation: QueryConsolidation = None, target: QueryTarget = None, value: IntoValue = None) -> Receiver:
        """
        Emits a query, which queryables with intersecting selectors will be able to reply to.

        The replies are provided to the given `handler` as instances of the `Reply` class.
        The `handler` can typically be a queue, a single callback or a pair of callbacks.
        The `handler`'s receiver is returned by the `get` function.

        :param selector: The selection of keys to query
        :param handler:
        :param consolidation: The consolidation to apply to replies
        :param target: The queryables that should be target to this query
        :param value: An optional value to attach to this query
        :return: The receiver of the handler
        :rtype: Receiver

        :Examples:

        Using a queue:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> for reply in s.get('key/expression', zenoh.Queue()):
        ...     try:
        ...         print(f"Received '{reply.ok.key_expr}': '{reply.ok.payload.decode('utf-8')}'")
        ...     except:
        ...         print(f"Received ERROR: '{reply.err.payload.decode('utf-8')}'")

        Using a single callback:

        >>> s.get('key/expression', lambda reply:
        ...     print(f"Received '{reply.ok.key_expr}': '{reply.ok.payload.decode('utf-8')}'")
        ...     if reply.ok is not None else print(f"Received ERROR: '{reply.err.payload.decode('utf-8')}'"))

        Using a reply callback and a termination callback:

        >>> s.get('key/expression', (
        ...     lambda reply:
        ...         print(f"Received '{reply.ok.key_expr}': '{reply.ok.payload.decode('utf-8')}'")
        ...         if reply.ok is not None else print(f"Received ERROR: '{reply.err.payload.decode('utf-8')}'"),
        ...     lambda:
        ...         print("No more replies")))
        """
        handler = Handler(handler, lambda x: Reply(x))
        kwargs = dict()
        if consolidation is not None:
            kwargs["consolidation"] = consolidation
        if target is not None:
            kwargs["target"] = target
        if value is not None:
            kwargs["value"] = Value(value)
        super().get(Selector(selector), handler.closure, **kwargs)
        return handler.receiver

    def declare_keyexpr(self, keyexpr: IntoKeyExpr) -> KeyExpr:
        """Informs Zenoh that you intend to use the provided Key Expression repeatedly.

        This function returns an optimized representation of the passed ``keyexpr``.

        It is generally not needed to declare key expressions, as declaring a subscriber, 
        a queryable, or a publisher will also inform Zenoh of your intent to use their
        key expressions repeatedly.
        """
        return KeyExpr(super().declare_keyexpr(KeyExpr(keyexpr)))

    def declare_queryable(self, keyexpr: IntoKeyExpr, handler: IntoHandler[Query, Any, Any], complete: bool = None):
        """Declares a queryable, which will receive queries intersecting with ``keyexpr``.

        These queries are passed to the `handler` as instances of the `Query` class.
        The `handler` can typically be a queue or a callback.
        The `handler`'s receiver is returned as the `receiver` field of the returned `Queryable`.
        The replies can be sent back by calling the `reply`function of the `Query`.

        :Examples:

        Using a callback:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> qabl = s.declare_queryable('key/expression', lambda query:
        ...     query.reply(zenoh.Sample('key/expression', 'value')))

        Using a queue:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> qabl = s.declare_queryable('key/expression', zenoh.Queue())
        >>> while True:
        ...     query = qabl.receiver.get()
        ...     query.reply(zenoh.Sample('key/expression', 'value'))
        ...     del query

        IMPORTANT: due to how RAII and Python work, you MUST bind this function's return value to a variable in order for it to function as expected.
        This is because as soon as a value is no longer referenced in Python, that value's destructor will run, which will undeclare your queryable, stopping it immediately.
        """
        handler = Handler(handler, lambda x: Query(x))
        kwargs = dict()
        if complete is not None:
            kwargs['complete'] = complete
        inner = super().declare_queryable(KeyExpr(keyexpr), handler.closure, **kwargs)
        return Queryable(inner, handler.receiver)

    def declare_publisher(self, keyexpr: IntoKeyExpr, priority: Priority = None, congestion_control: CongestionControl = None):
        """
        Declares a publisher, which may be used to send values repeatedly onto a same key expression.

        Written resources that match the given key will only be sent on the network
        if matching subscribers exist in the system.

        :param keyexpr: The key expression to publish to
        :param priority: The priority to use when routing the published data
        :param congestion_control: The congestion control to use when routing the published data
        :rtype: Publisher

        :Examples:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> pub = s.declare_publisher('key/expression')
        >>> pub.put('value')
        """
        kwargs = dict()
        if priority is not None:
            kwargs['priority'] = priority
        if congestion_control is not None:
            kwargs['congestion_control'] = congestion_control
        return Publisher(super().declare_publisher(KeyExpr(keyexpr), **kwargs))

    def declare_subscriber(self, keyexpr: IntoKeyExpr, handler: IntoHandler[Sample, Any, Any], reliability: Reliability = None) -> Subscriber:
        """
        Declares a subscriber, which will receive any published sample with a key expression intersecting ``keyexpr``.

        These samples are provided to the `handler` as instances of the `Sample` class.
        The `handler` can typically be a queue or a callback.
        The `handler`'s receiver is returned as the `receiver` field of the returned `Subscriber`.

        :param keyexpr: The key expression to subscribe to
        :param handler:
        :param reliability: the reliability to use when routing the subscribed samples
        :rtype: Subscriber

        :Examples:

        Using a callback:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> sub = s.declare_subscriber('key/expression', lambda sample:
        ...     print(f"Received '{sample.key_expr}': '{sample.payload.decode('utf-8')}'")

        Using a queue:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> sub = s.declare_subscriber('key/expression', zenoh.Queue())
        >>> for sample in sub.receiver:
        >>>     print(f"{sample.key_expr}: {sample.payload.decode('utf-8')}")

        IMPORTANT: due to how RAII and Python work, you MUST bind this function's return value to a variable in order for it to function as expected.
        This is because as soon as a value is no longer referenced in Python, that value's destructor will run, which will undeclare your subscriber, deactivating the subscription immediately.
        """
        handler = Handler(handler, lambda x: Sample._upgrade_(x))
        kwargs = dict()
        if reliability is not None:
            kwargs['reliability'] = reliability
        s = super().declare_subscriber(KeyExpr(keyexpr), handler.closure, **kwargs)
        return Subscriber(s, handler.receiver)

    def declare_pull_subscriber(self, keyexpr: IntoKeyExpr, handler: IntoHandler[Sample, Any, Any], reliability: Reliability = None) -> PullSubscriber:
        """
        Declares a pull-mode subscriber, which will receive a single published sample with a key expression intersecting ``keyexpr`` any time its ``pull`` method is called.

        These samples are passed to the `handler`'s closure as instances of the `Sample` class.
        The `handler` can typically be a queue or a callback.
        The `handler`'s receiver is returned as the `receiver` field of the returned `PullSubscriber`.

        :param keyexpr: The key expression to subscribe to
        :param handler:
        :param reliability: the reliability to use when routing the subscribed samples
        :rtype: PullSubscriber

        :Examples:

        >>> import zenoh
        >>> s = zenoh.open({})
        >>> sub = s.declare_pull_subscriber('key/expression', lambda sample:
        ...     print(f"Received '{sample.key_expr}': '{sample.payload.decode('utf-8')}'"))
        ...
        >>> sub.pull()
        """
        handler = Handler(handler, lambda x: Sample._upgrade_(x))
        kwargs = dict()
        if reliability is not None:
            kwargs['reliability'] = reliability
        s = super().declare_pull_subscriber(KeyExpr(keyexpr), handler.closure, **kwargs)
        return PullSubscriber(s, handler.receiver)

    def close(self):
        """Attempts to close the Session.
        
        The session will only be closed if all publishers, subscribers and queryables based on it have been undeclared, and there are no more python references to it.
        """
        pass

    def info(self):
        "Returns an accessor for informations about this Session"
        return Info(self)


class Info:
    def __init__(self, session: _Session):
        self.session = session

    def zid(self) -> ZenohId:
        "Returns this Zenoh Session's identifier"
        return ZenohId._upgrade_(self.session.zid())

    def routers_zid(self) -> List[ZenohId]:
        "Returns the neighbooring routers' identifiers"
        return [ZenohId._upgrade_(id) for id in self.session.routers_zid()]

    def peers_zid(self) -> List[ZenohId]:
        "Returns the neighbooring peers' identifiers"
        return [ZenohId._upgrade_(id) for id in self.session.peers_zid()]



#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
import abc
from typing import Generic, Callable, Union, Any, TypeVar, Tuple, List
from threading import Condition, Thread
from collections import deque
import time

from .zenoh import _Queue

In = TypeVar("In")
Out = TypeVar("Out")
Receiver = TypeVar("Receiver")
CallbackCall = Callable[[In], Out]
CallbackDrop = Callable[[], None]

class IClosure(Generic[In, Out]):
    """
    A Closure is a pair of a ``call`` function that will be used as a callback,
    and a ``drop`` function that will be called when the closure is destroyed.
    """
    @property
    @abc.abstractmethod
    def call(self) -> Callable[[In], Out]:
        """
        Returns the closure's call function as a lambda.
        """
        ...
    @property
    @abc.abstractmethod
    def drop(self) -> Callable[[], None]:
        """
        Returns the closure's destructor as a lambda.
        """
        ...
    def __enter__(self):
        drop = self.drop
        if drop is not None:
            drop()
    def __exit__(self, *args):
        drop = self.drop
        if drop is not None:
            drop()

class IHandler(Generic[In, Out, Receiver]):
    """
    A Handler is a value that may be converted into a callback closure for zenoh to use on one side, while possibly providing a receiver for the data that zenoh would provide through that callback.
    """
    @property
    @abc.abstractmethod
    def closure(self) -> IClosure[In, Out]:
        """
        The part of the handler that should be passed as a callback to a zenoh function.
        """
        ...
    @property
    @abc.abstractmethod
    def receiver(self) -> Receiver:
        "The part of the handler that should be used as the receiver when the handler is channel-like."
        ...

IntoClosure = Union[IHandler[In, Out, Any], IClosure[In, Out], Tuple[CallbackCall, CallbackDrop], CallbackCall]
class Closure(IClosure, Generic[In, Out]):
    """
    A Closure is a pair of a ``call`` function that will be used as a callback,
    and a ``drop`` function that will be called when the closure is destroyed.
    """
    def __init__(self, closure: IntoClosure[In, Out], type_adaptor: Callable[[Any], In] = None, prevent_direct_calls=False):
        _call_ = None
        _drop_ = lambda: None
        if isinstance(closure, IHandler):
            closure = closure.closure
            # dev-note: do not elif here, the  next if will catch the obtained closure.
        if isinstance(closure, IClosure):
            _call_ = closure.call
            _drop_ = closure.drop
        elif isinstance(closure, tuple):
            _call_, _drop_ = closure
        elif callable(closure):
            _call_ = closure
        else:
            raise TypeError("Unexpected type as input for zenoh.Closure")
        if type_adaptor is not None:
            adapted = lambda *args: _call_(type_adaptor(*args))
        else:
            adapted = _call_
        if prevent_direct_calls:
            queue = Queue(128)
            def readqueue():
                for x in queue:
                    adapted(*x)
                    x = None
            t = Thread(target=readqueue)
            t.start()
            self._call_ = lambda *args: queue.put(args)
            def drop():
                queue.close()
                t.join()
                _drop_()
            self._drop_ = drop
        else:
            self._call_ = _call_
            self._drop_ = _drop_

    @property
    def call(self) -> Callable[[In], Out]:
        return self._call_

    @property
    def drop(self) -> Callable[[], None]:
        return self._drop_

IntoHandler = Union[IHandler[In, Out, Receiver], IClosure[In, Out],  Tuple[IClosure, Receiver], Tuple[CallbackCall,CallbackDrop, Receiver], Tuple[CallbackCall,CallbackDrop], CallbackCall]
class Handler(IHandler, Generic[In, Out, Receiver]):
    """
    A Handler is a value that may be converted into a callback closure for zenoh to use on one side, while possibly providing a receiver for the data that zenoh would provide through that callback.

    Note that the values will be piped onto a ``Queue`` before being sent to your handler by another Thread unless either:
        a) ``input`` is already an instance of ``Closure`` or ``Handler`` where ``input.closure`` is an instance of ``Closure``
        b) ``prevent_direct_calls`` is set to ``False``
    """
    def __init__(self, input: IntoHandler[In, Out, Receiver], type_adaptor: Callable[[Any], In] = None, prevent_direct_calls = True):
        self._receiver_ = None
        if isinstance(input, IHandler):
            self._receiver_ = input.receiver
            self._closure_ = input.closure
        elif isinstance(input, IClosure):
            self._closure_ = input
        elif isinstance(input, tuple):
            if isinstance(input[0], IClosure):
                self._closure_, self._receiver_ = input
            elif len(input) == 3:
                call, drop, self._receiver_ = input
                self._closure_ = (call, drop)
            else:
                self._closure_ = input
        else:
            self._closure_ = input
        self._closure_ = Closure(self._closure_, type_adaptor, prevent_direct_calls and not isinstance(self._closure_, Closure))

    @property
    def closure(self) -> IClosure[In, Out]:
        return self._closure_
    @property
    def receiver(self) -> Receiver:
        return self._receiver_



class ListCollector(IHandler[In, None, Callable[[],List[In]]], Generic[In]):
    """
    A simple collector that aggregates values into a list.

    When used as a handler, it provides a callback that appends elements to a list,
    and provides a function that will await the closing of the callback before returning said list.
    """
    def __init__(self, timeout=None):
        self._vec_ = []
        self._cv_ = Condition()
        self._done_ = False
        self.timeout = timeout
    
    @property
    def closure(self):
        def call(x):
            self._vec_.append(x)
        def drop():
            with self._cv_:
                self._done_ = True
                self._cv_.notify()
        return Closure((call, drop))
    
    @property
    def receiver(self):
        def wait():
            with self._cv_:
                if not self._done_:
                    self._cv_.wait(timeout=self.timeout)
                return self._vec_
        return wait

class Queue(IHandler[In, None, 'Queue'], Generic[In]):
    """
    A binding for a Rust multi-producer, single-consumer queue implementation.

    When used as a handler, it provides itself as the receiver, and will provide a
    callback that appends elements to the queue.

    Can be bounded by passing a maximum size as ``bound``.
    """
    def __init__(self, bound: int = None):
        self._inner_ = _Queue(bound)
    
    @property
    def closure(self) -> IClosure[In, None]:
        def call(x): self.put(x)
        def drop(): self.close()
        return Closure((call, drop))
    
    @property
    def receiver(self) -> 'Queue':
        return self
    
    def put(self, value):
        """
        Puts one element on the queue.

        Raises a ``PyBrokenPipeError`` if the Queue has been closed.
        """
        return self._inner_.put(value)


    def get(self, timeout: float = None):
        """
        Gets one element from the queue.

        Raises a ``StopIteration`` exception if the queue was closed before the timeout ran out,
        this allows using the Queue as an iterator in for-loops.
        Raises a ``TimeoutError`` if the timeout ran out.
        """
        return self._inner_.get(timeout)
    
    def close(self):
        return self._inner_.close()
    
    def get_remaining(self, timeout: float = None) -> List[In]:
        """
        Awaits the closing of the queue, returning the remaining queued values in a list.
        The values inserted into the queue up until this happens will be available through ``get``.

        Raises a ``TimeoutError`` if the timeout in seconds provided was exceeded before closing,
        whose ``args[0]`` will contain the elements that were collected before timing out.
        """
        return self._inner_.get_remaining()

    def __iter__(self):
        return self
    def __next__(self):
        return self.get()

if __name__ == "__main__":
    def get(collector):
        import time
        def target():
            with Closure(collector) as closure:
                closure = Closure(collector)
                closure.call('hi')
                closure.call('there')
        Thread(target=target).start()

    collector = ListCollector()
    get(collector)
    print(collector.receiver())
    assert collector.receiver() == ["hi", "there"]

    print("done")





#
# Copyright (c) 2022 ZettaScale Technology
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors:
#   ZettaScale Zenoh Team, <zenoh@zettascale.tech>
#
import abc
from typing import Union, Tuple, Optional, List
import json

from .enums import Encoding, SampleKind
from .zenoh import _Value, _Encoding, _Sample, _SampleKind, _Reply, _ZenohId, _Timestamp, _Hello
from .keyexpr import KeyExpr, IntoKeyExpr

class IValue:
    "The IValue interface exposes how to recover a value's payload in a binary-serialized format, as well as that format's encoding."
    @property
    @abc.abstractmethod
    def payload(self) -> bytes:
        "The value itself, as an array of bytes"
        ...
    
    @property
    @abc.abstractmethod
    def encoding(self) -> Encoding:
        "The value's encoding"
        ...

IntoValue = Union[IValue, bytes, str, int, float, object]

class Value(_Value, IValue):
    """
    A Value is a pair of a binary payload, and a mime-type-like encoding string.
    
    When constructed with ``encoding==None``, the encoding will be selected depending on the payload's type.
    """
    def __new__(cls, payload: IntoValue, encoding: Encoding=None):
        if encoding is None:
            if isinstance(payload, Value):
                return payload
            return Value.autoencode(payload)
        else:
            if not isinstance(payload, bytes):
                raise TypeError("`encoding` was passed, but `payload` is not of type `bytes`")
            return Value.new(payload, encoding)
    
    @staticmethod
    def autoencode(value: IntoValue) -> 'Value':
        "Automatically encodes the value based on its type"
        if isinstance(value, IValue):
            return Value.new(value.payload, value.encoding)
        if isinstance(value, bytes):
            return Value.new(value, Encoding.APP_OCTET_STREAM())
        if isinstance(value, str):
            return Value.new(value.encode(), Encoding.TEXT_PLAIN())
        if isinstance(value, int):
            return Value.new(f"{value}".encode(), Encoding.APP_INTEGER())
        if isinstance(value, float):
            return Value.new(f"{value}".encode(), Encoding.APP_FLOAT())
        return Value.new(json.dumps(value).encode(), Encoding.APP_JSON())
    
    @staticmethod
    def new(payload: bytes, encoding: Encoding = None) -> 'Value':
        return Value._upgrade_(_Value.new(payload, encoding))

    @property
    def payload(self) -> bytes:
        return super().payload

    @payload.setter
    def payload(self, payload: bytes):
        super().with_payload(payload)

    @property
    def encoding(self) -> Encoding:
        return Encoding(super().encoding)

    @encoding.setter
    def encoding(self, encoding: Encoding):
        super().with_encoding(encoding)

    @staticmethod
    def _upgrade_(inner: _Value) -> 'Value':
        if inner is None:
            return None
        if isinstance(inner, Value):
            return inner
        return _Value.__new__(Value, inner)

class ZenohId(_ZenohId):
    """A Zenoh UUID"""
    @staticmethod
    def _upgrade_(this: _ZenohId) -> 'ZenohId':
        return _ZenohId.__new__(ZenohId, this)
    def __str__(self) -> str:
        return super().__str__()
    def __repr__(self) -> str:
        return str(self)

class Timestamp(_Timestamp):
    """
    A timestamp taken from the Zenoh HLC (Hybrid Logical Clock).

    These timestamps are guaranteed to be unique, as each machine annotates its perceived time with a UUID, which is used as the least significant part of the comparison operation.
    """
    @staticmethod
    def _upgrade_(this: _Timestamp) -> 'Timestamp':
        return _Timestamp.__new__(Timestamp, this)
    @property
    def get_time(self) -> int:
        """
        Returns the time part, as generated by the Zenoh HLC in NTP64 format (See https://datatracker.ietf.org/doc/html/rfc5905#section-6).
        """
        return super().time
    @property
    def seconds_since_unix_epoch(self) -> float:
        """
        Returns the number of seconds since the Unix Epoch.

        Considering the large number of seconds since the Unix Epoch, the precision of the resulting f64 is in the order of microseconds.
        Therefore, it should not be used for comparison. Directly comparing Timestamp objects is preferable.
        """
        return super().seconds_since_unix_epoch


IntoSample = Union[_Sample, Tuple[IntoKeyExpr, IntoValue, SampleKind], Tuple[KeyExpr, IntoValue]]
class Sample(_Sample):
    """
    A KeyExpr-Value pair, annotated with the kind (PUT or DELETE) of publication used to emit it and a timestamp.
    """
    def __new__(cls, key: IntoKeyExpr, value: IntoValue, kind: SampleKind = None, timestamp: Timestamp = None):
        kind = _SampleKind.PUT if kind is None else kind
        return Sample._upgrade_(super().new(KeyExpr(key), Value(value), kind, timestamp))
    @property
    def key_expr(self) -> KeyExpr:
        "The sample's key expression"
        return KeyExpr(super().key_expr)
    @property
    def value(self) -> Value:
        "The sample's value"
        return Value._upgrade_(super().value)
    @property
    def payload(self) -> bytes:
        "A shortcut to ``self.value.payload``"
        return super().payload
    @property
    def encoding(self) -> Encoding:
        "A shortcut to ``self.value.encoding``"
        return Encoding(super().encoding)
    @property
    def kind(self) -> SampleKind:
        "The sample's kind"
        return SampleKind(super().kind)
    @property
    def timestamp(self) -> Optional[Timestamp]:
        "The sample's  timestamp. May be None."
        ts = super().timestamp
        return None if ts is None else Timestamp._upgrade_(ts)
    @staticmethod
    def _upgrade_(inner: _Sample) -> 'Sample':
        if isinstance(inner, Sample):
            return inner
        return _Sample.__new__(Sample, inner)

class Reply(_Reply):
    """
    A reply to a query (``Session.get``).
    
    A single query can result in multiple replies from multiple queryables.
    """
    def __new__(cls, inner: _Reply):
        return super().__new__(cls, inner)
    @property
    def replier_id(self) -> ZenohId:
        "The reply's sender's id."
        return ZenohId._upgrade_(super().replier_id)
    
    @property
    def is_ok(self) -> bool:
        """
        Checks if the reply is `ok`.

        Returns `True` if the reply is `ok`, `False` otherwise
        """
        return super.is_ok()

    @property
    def ok(self) -> Sample:
        """
        The reply's inner data sample.

        Raises a ``ZError`` if the ``self`` is actually an ``err`` reply.
        """
        return Sample._upgrade_(super().ok)
    @property
    def err(self) -> Value:
        """
        The reply's error value.

        Raises a ``ZError`` if the ``self`` is actually an ``ok`` reply.
        """
        return Value._upgrade_(super().err)

class Hello(_Hello):
    "Represents a single Zenoh node discovered through scouting."
    @property
    def zid(self) -> ZenohId:
        "The node's Zenoh UUID."
        zid = super().zid
        return None if zid is None else ZenohId._upgrade_(zid)
    @property
    def whatami(self) -> str:
        "The node's type, returning either None, 'peer', 'router', or 'client'."
        return super().whatami
    @property
    def locators(self) -> List[str]:
        "The locators through which this node may be adressed."
        return super().locators
    @staticmethod
    def _upgrade_(inner: _Hello) -> 'Sample':
        if isinstance(inner, Hello):
            return inner
        return _Hello.__new__(Hello, inner)
    def __str__(self):
        return super().__str__()